module.exports =
    PUBSUB: "http://jabber.org/protocol/pubsub"
    PUBSUB_EVENT: "http://jabber.org/protocol/pubsub#event"
    PUBSUB_OWNER: "http://jabber.org/protocol/pubsub#owner"
    PUBSUB_NODE_CONFIG: "http://jabber.org/protocol/pubsub#node_config"
    PUBSUB_META_DATA: "http://jabber.org/protocol/pubsub#meta-data"
    PUBSUB_SUBSCRIBE_AUTHORIZATION: "http://jabber.org/protocol/pubsub#subscribe_authorization"
    DISCO_INFO: "http://jabber.org/protocol/disco#info"
    DISCO_ITEMS: "http://jabber.org/protocol/disco#items"
    DATA: "jabber:x:data"
    REGISTER: "jabber:iq:register"
    COMMANDS: "http://jabber.org/protocol/commands"
    RSM: "http://jabber.org/protocol/rsm"
    BUDDYCLOUD_V1: "http://buddycloud.org/v1"
    MAM: "urn:xmpp:mam:tmp"
    FORWARD: "urn:xmpp:forward:0"
    VERSION: "jabber:iq:version"
